/***************************************************************************
                          qfrpostscript.cpp
                         ----------------------------
     begin                : 2005-06-23
     copyright            : (C) 2005 by Serghei Amelian
     email                : serghei.amelian@gmail.com
 ***************************************************************************/
#include <math.h>
#include <qdatetime.h>
#include <qimage.h>

#include "../config.h"
#include "qfrtiffio.h"

#include "qfrpostscript.h"


static const char *sBeginDocument =
  "%%!PS-Adobe-3.0\n"
  "%%%%Creator: QFaxReader " VERSION "\n"
  "%%%%Title: %s\n"
  "%%%%CreationDate: %s\n"
  "%%%%DocumentData: Clean7Bit\n"
  "%%%%Origin: 0 0\n"
  "%%%%BoundingBox: 0 0 %d %d\n"
  "%%%%Orientation: %s\n"
  "%%%%LanguageLevel: %d\n"
  "%%%%Pages: (atend)\n"
  "%%%%EndComments\n"
  "%%%%BeginSetup\n"
  "%s"
  "%%%%EndSetup\n";


static const char *sSetup =
  "gsave newpath clippath pathbbox grestore\n"
  "  4 2 roll 2 copy translate\n"
  "  exch 3 1 roll sub 3 1 roll sub exch\n"
  "  currentpagedevice /PageSize get aload pop\n"
  "  exch 3 1 roll div 3 1 roll div abs exch abs\n"
  "  2 copy gt { exch } if pop\n"
  "  dup 1 lt { dup scale } { pop } ifelse\n";


static const char *sEndDocument =
  "%%%%Trailer\n"
  "%%%%Pages: %d\n"
  "%%%%EOF\n";


static const char *sBeginPage =
  "%%%%Page: %d %d\n"
  "gsave\n"
  "%s"
  "100 dict begin\n";


static const char *sEndPage =
  "grestore\n"
  "showpage\n";


static const char *sRectangle =
  "newpath\n"
  "%d %d moveto\n"
  "%d 0 rlineto\n"
  "0 -%d rlineto\n"
  "-%d 0 rlineto\n"
  "0 %d rlineto\n"
  ".1 setlinewidth\n"
  "stroke\n";


static const char *sBeginImage =
  "gsave\n"
  "%d %d translate\n"
  "%d %d scale\n"
  "0 1 translate\n";


static const char *sBeginBWImage1 =
  "%%ImageData: %d %d 1 1 0 1 2 \"image\"\n"
  "/DataString %d string def\n"
  "%d %d %d [ %d 0 0 -%d 0 %d ]\n"
  "{ currentfile DataString readhexstring pop } bind image\n";

static const char *sBeginColorImage1 =
  "/bwproc {\n"
  "    rgbproc\n"
  "    dup length 3 idiv string 0 3 0\n"
  "    5 -1 roll {\n"
  "        add 2 1 roll 1 sub dup 0 eq {\n"
  "            pop 3 idiv\n"
  "            3 -1 roll\n"
  "            dup 4 -1 roll\n"
  "            dup 3 1 roll\n"
  "            5 -1 roll put\n"
  "            1 add 3 0\n"
  "        } { 2 1 roll } ifelse\n"
  "    } forall\n"
  "    pop pop pop\n"
  "} def\n"
  "/colorimage where {pop} {\n"
  "    /colorimage {pop pop /rgbproc exch def {bwproc} image} bind def\n"
  "} ifelse\n"
  "%ImageData: 808 630 8 3 0 1 2 \"false 3 colorimage\"\n"
  "/line 2424 string def\n"
  "808 630 8\n"
  "[808 0 0 -630 0 630]\n"
  "{currentfile line readhexstring pop} bind\n"
  "false 3 colorimage\n";


static const char *sBeginImage2 =
  "/DeviceGray setcolorspace\n"
  "{ %% exec\n"
   "/im_stream currentfile /ASCII85Decode filter def\n"
   "<<\n"
   "/ImageType 1\n"
   "/Width %d\n"
   "/Height %d\n"
   "/ImageMatrix [ %d 0 0 -%d 0 %d ]\n"
   "/BitsPerComponent 1\n"
   "/Interpolate true\n"
   "/Decode [0 1]\n"
   "/DataSource im_stream\n"
   "     <<\n"
   "      /EndOfLine true\n"
   "      /EndOfBlock false\n"
   "      /EncodedByteAlign true\n"
   "      /Rows %d\n"
   "     >> /CCITTFaxDecode filter\n"
   ">> image\n"
   "im_stream status { im_stream flushfile } if\n"
   "}\n"
   "exec\n";

static const char *sEndImage =
  "grestore\n";


QfrPostscript::QfrPostscript(FILE *f)
: f(f), pages(0)
{
}


QfrPostscript::~QfrPostscript()
{
}


bool QfrPostscript::beginDocument(const char *title, int pagesPerSheet, int width, int height, int top, int bottom, int left, int right, int level, bool border)
{
  spacing = 6;
  m_pagesPerSheet = pagesPerSheet;
  m_border = border;
  pages = images = 0;
  m_width = width;
  m_height = height;
  m_top = top;
  m_bottom = bottom;
  m_left = left;
  m_right = right;
  m_level = level;
  m_landscape = (2 == pagesPerSheet);
  return 0 <= fprintf(f, sBeginDocument,
    title,
    QDateTime::currentDateTime().toString().latin1(),
    width,
    height,
    m_landscape ? "Landscape" : "Portrait",
    level,
    1 == level ? "" : sSetup);
}


bool QfrPostscript::endDocument()
{
  QString s1, s2;
  if(pages) s1 = sEndPage;
  s2.sprintf(sEndDocument, pages);
  PUTS(s1 + s2);
  return true;
}


bool QfrPostscript::newPage()
{
  QString s1, s2;
  if(pages) s1 = sEndPage;
  pages++;
  s2.sprintf(sBeginPage, pages, pages, m_landscape ? "90 rotate\n" : "");
  PUTS(s1 + s2);
  return true;
}


bool QfrPostscript::drawRectangle(int x, int y, int w, int h)
{
  QString s1;
  s1.sprintf(sRectangle, x + m_left, (m_landscape ? 0 : m_height) - m_top - y, w, h, w, h);
  PUTS(s1);
  return true;
}


bool QfrPostscript::pushImage(QfrTiffIO &tif)
{
  if(images >= m_pagesPerSheet)
    images = 0;
  if(!images)
    newPage();

  int x = images % 2 * availWidth();
  int y = ((images & 2) >> 1) * availHeight();
  int w = availWidth();
  int h = availHeight();

  if(x) x += spacing;
  if(y) y += spacing;

  if(1 == m_level)
    if(!drawLevel1Image(x, y, w, h, tif))
      return false;
  else
    if(!drawLevel2Image(x, y, w, h, tif))
      return false;

  if(m_border)
    if(!drawRectangle(x, y, w, h))
      return false;

  images++;
  return true;
}


int QfrPostscript::availWidth() const
{
  int totalWidth = (m_landscape ? m_height : m_width) - m_left - m_right - (m_pagesPerSheet == 1 ? 0 : spacing);
  return totalWidth / (m_pagesPerSheet == 1 ? 1  : 2);
}


int QfrPostscript::availHeight() const
{
  int totalHeight = (m_landscape ? m_width : m_height) - m_top - m_bottom - (m_pagesPerSheet == 1 ? 0 : spacing);
  return totalHeight / (m_pagesPerSheet == 4 ? 2  : 1);
}


bool QfrPostscript::drawLevel1Image(int x, int y, int w, int h, QfrTiffIO &tif)
{
  x += m_left;
  y += m_top;

  uint32 iw, ih;
  tif.GetField(TIFFTAG_IMAGEWIDTH, &iw);
  tif.GetField(TIFFTAG_IMAGELENGTH, &ih);

  float xres, yres;
  if(!tif.GetField(TIFFTAG_XRESOLUTION, &xres)) xres = 300.;
  if(!tif.GetField(TIFFTAG_YRESOLUTION, &yres)) yres = 300.;

  int nw = (int)round((iw / xres) * 72);
  int nh = (int)round((ih / yres) * 72);
  float ratio = (float)nh / (float)nw;

  if(nw > w)
    {
      nw = w;
      nh = (uint32)round(nw * ratio);
    }

  if(nh > h)
    {
      nh = h;
      nw = (uint32)round(nh / ratio);
    }

  w = nw;
  h = nh;

  // photometric interpretation
  short pmi = 0;
  tif.GetField(TIFFTAG_PHOTOMETRIC, &pmi);

  // bits per sample
  short bits = 1;
  tif.GetField(TIFFTAG_BITSPERSAMPLE, &bits);

  // samples per pixel
  short samples = 1;
  tif.GetField(TIFFTAG_SAMPLESPERPIXEL, &samples);

  if(3 > pmi && 1 == samples)
    return drawLevel1BWImage(x, y, w, h, iw, ih, bits, 0 == pmi, tif);
  return drawLevel1ColorImage(x, y, w, h, iw, ih, tif);
}


bool QfrPostscript::drawLevel1BWImage(int x, int y, int w, int h, int iw, int ih, int bits, bool minIsWhite, QfrTiffIO &tif)
{
  tsize_t len = tif.ScanlineSize();

  QString s1, s2;
  s1.sprintf(sBeginImage, x, (m_landscape ? 0 : m_height) - y - h * 2, w, h);
  s2.sprintf(sBeginBWImage1, iw, ih, len, iw, ih, bits, iw, ih, ih);

  PUTS(s1 + s2);

  QfrAsciiHexEncoder enc(f);
  enc.begin();
  uchar *buf = new uchar[len];
  for(int row = 0; row < ih; row++)
    {
      tif.ReadScanline(buf, row);
      if(minIsWhite)
        for(register int i = 0; i < len; i++)
          buf[i] = ~buf[i];
      enc.encode(buf, len);
    }
  delete [] buf;
  enc.end();

  PUTS(sEndImage);

  return true;

}


bool QfrPostscript::drawLevel1ColorImage(int x, int y, int w, int h, int iw, int ih, QfrTiffIO &tif)
{
  QString s1, s2;
  s1.sprintf(sBeginImage, x, (m_landscape ? 0 : m_height) - y - h * 2, w, h);
  s2.sprintf(sBeginColorImage1);

  PUTS(s1 + s2);

  QImage image(iw, ih, 32);
  //tif.ReadRGBAImageOriented(iw, ih, (uint32*)image.bits(), ORIENTATION_TOPLEFT);
  tif.ReadRGBAImage(iw, ih, (uint32*)image.bits());
  //image = image.swapRGB();

  QfrAsciiHexEncoder enc(f);
  enc.begin();
  enc.encode(image.bits(), iw * ih * sizeof(uint32));
  enc.end();

  PUTS(sEndImage);

  return true;
}


bool QfrPostscript::drawLevel2Image(int x, int y, int w, int h, QfrTiffIO &tif)
{
  x += m_left;
  y += m_top;

  uint32 iw, ih;
  tif.GetField(TIFFTAG_IMAGEWIDTH, &iw);
  tif.GetField(TIFFTAG_IMAGELENGTH, &ih);

  float xres, yres;
  if(!tif.GetField(TIFFTAG_XRESOLUTION, &xres)) xres = 300.;
  if(!tif.GetField(TIFFTAG_YRESOLUTION, &yres)) yres = 300.;

  uint32 nw = (int)round((iw / xres) * 72);
  uint32 nh = (int)round((ih / yres) * 72);
  float ratio = (float)nh / (float)nw;

  if(nw > (uint32)w)
    {
      nw = w;
      nh = (uint32)round(nw * ratio);
    }

  if(nh > (uint32)h)
    {
      nh = h;
      nw = (uint32)round(nh / ratio);
    }

  w = nw;
  h = nh;

  QString s1, s2;
  s1.sprintf(sBeginImage, x, (m_landscape ? 0 : m_height) - y - h * 2, w, h);
  s2.sprintf(sBeginImage2, iw, ih, iw, ih, ih, ih);
  PUTS(s1 + s2);

  QfrAscii85Encoder enc(f);
  enc.begin();

  tdata_t buf;
  tstrip_t strip;
  uint32 *bc;
  uint32 stripsize;

  tif.GetField(TIFFTAG_STRIPBYTECOUNTS, &bc);
  stripsize = bc[0];
  buf = _TIFFmalloc(stripsize);
  for(strip = 0; strip < tif.NumberOfStrips(); strip++)
    {
      if(bc[strip] > stripsize)
        {
          buf = _TIFFrealloc(buf, bc[strip]);
          stripsize = bc[strip];
        }
      tif.ReadRawStrip(strip, buf, bc[strip]);
      enc.encode((const uchar*)buf, bc[strip]);
    }
  _TIFFfree(buf);

  enc.end();

  PUTS(sEndImage);
  return true;
}
